program Oper

  use sep
  use fbi_plane_mod
  use tomo_mod

  implicit none

  integer                             :: nt, nte, n1, n2, wfld_j, gpu_n, gpu_o
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: bc1, bc2, bc3, bc4, np
  real                                :: dt, d1, o1, d2, o2, op, dp
  real                                :: vmin, vmax, ssmin, ssmax, eps

  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:,:,:), allocatable :: vel_init, vel0, slowsq0, slowsq1, gradient0, gradient, gradient_m, search
  real, dimension(:,:,:), allocatable :: data_obs, data

  integer                             :: stat, iter, niter, itry, ntry, counter

  double precision                    :: rd, rm, r0, r1, r2, r3, step_size, beta

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)
  call from_history("n3", np)
  call from_history("o3", op)
  call from_history("d3", dp)

  ! Read wavelet information
  call from_aux("wavelet", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)

  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param('niter', niter, 10)
  call from_param('ntry', ntry, 2)
  call from_param('eps', eps)

  call from_param('vmin', vmin)
  call from_param('vmax', vmax)

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, gpu_n, gpu_o, np, op, dp)
  call init4(nte)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", np)
  call to_history("o3", op)
  call to_history("d3", dp)

  if(exist_file("mmov")) then
    call to_history("n1", n1, "mmov")
    call to_history("o1", o1, "mmov")
    call to_history("d1", d1, "mmov")
    call to_history("n2", n2, "mmov")
    call to_history("o2", o2, "mmov")
    call to_history("d2", d2, "mmov")
    call to_history("n3", np, "mmov")
    call to_history("o3", op, "mmov")
    call to_history("d3", dp, "mmov")
  endif

  if(exist_file("dmmov")) then
    call to_history("n1", n1, "dmmov")
    call to_history("o1", o1, "dmmov")
    call to_history("d1", d1, "dmmov")
    call to_history("n2", n2, "dmmov")
    call to_history("o2", o2, "dmmov")
    call to_history("d2", d2, "dmmov")
    call to_history("n3", np, "dmmov")
    call to_history("o3", op, "dmmov")
    call to_history("d3", dp, "dmmov")
  endif

  call to_history("n1", 3, "res")
  call to_history("o1", 0., "res")
  call to_history("d1", 1., "res")

  ! Allocation
  allocate( wavelet(nt))
  allocate( res(3))
  allocate( vel0(n1, n2, np))
  allocate( vel_init(n1, n2, np))
  allocate( slowsq0(n1, n2, np))
  allocate( slowsq1(n1, n2, np))
  allocate( gradient0(n1, n2, np))
  allocate( gradient(n1, n2, np))
  allocate( gradient_m(n1, n2, np))
  allocate( search(n1, n2, np))
  allocate( data(nte, rec_nx, np))
  allocate( data_obs(nte, rec_nx, np))

  write(0,*) "allocated (3d in inv only) GB", nte/1024.*rec_nx/1024.*np/1024.*4.*2.

  wavelet = 0.
  vel0 = 0.
  vel_init = 0.
  data_obs = 0.
  res = 0.
  step_size = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")
  call sep_read(data_obs, "data")

  where(vel0 < vmin) vel0 = vmin
  where(vel0 > vmax) vel0 = vmax

  ssmax = 1./(vmin*vmin)
  ssmin = 1./(vmax*vmax)

  vel_init = vel0
  slowsq0 = 1./(vel0*vel0)

  call init_alloc(wavelet)

  counter = 0

  do iter=1,niter
    write(0,*) " "
    call model_update(vel0)
    call nlL_oper2(.false., data)
    data = data - data_obs
    rd = sum(dprod(data,data))
    gradient_m(:,:,1) = 0.
    gradient_m(:,:,2:np) = slowsq0(:,:,2:np) - slowsq0(:,:,1:np-1)
    rm = eps*eps*sum(dprod(gradient_m,gradient_m))
    r0 = rd + rm

    res(1) = rd
    res(2) = rm
    res(3) = r0
    write(0,*) "res =", res

    gradient0 = gradient

    call L_oper2(.true., .false., gradient, data)
    gradient_m(:,:,1) = slowsq0(:,:,1) - slowsq0(:,:,2)
    gradient_m(:,:,2:np-1) = 2.*slowsq0(:,:,2:np-1) - slowsq0(:,:,1:np-2) - slowsq0(:,:,3:np)
    gradient_m(:,:,np) = slowsq0(:,:,np) - slowsq0(:,:,np-1)
    gradient_m = eps*eps*gradient_m
    gradient = gradient + gradient_m

    if(iter .eq. 1) then
      search = gradient
    else
      beta = sum(dprod(gradient,gradient-gradient0))/sum(dprod(gradient0,gradient0))
      write(0,*) "beta =", beta
      beta = max(beta,0.)
      search = gradient + beta * search
    end if

    write(0,*) "grad stats", minval(gradient), maxval(gradient), sum(gradient)/size(gradient), sqrt(sum(gradient*gradient))
    write(0,*) "gram stats", minval(gradient_m), maxval(gradient_m), sum(gradient_m)/size(gradient_m), sqrt(sum(gradient_m*gradient_m))
    write(0,*) "srch stats", minval(search), maxval(search), sum(search)/size(search), sqrt(sum(search*search))
    write(0,*) " "

    do itry=1,2*ntry

      if(itry .eq. ntry .or. step_size .eq. 0.) then
        step_size = sqrt(sum(slowsq0*slowsq0)) / sqrt(sum(search*search)) / 100.
        write(0,*) "Step size is resetted to ", step_size
      endif

      slowsq1 = slowsq0 - search * step_size * 0.5
      where(slowsq1 < ssmin) slowsq1 = ssmin
      where(slowsq1 > ssmax) slowsq1 = ssmax
      vel0 = 1./sqrt(slowsq1)
      call model_update(vel0)
      call nlL_oper2(.false., data)
      data = data - data_obs
      rd = sum(dprod(data,data))
      gradient_m(:,:,1) = 0.
      gradient_m(:,:,2:np) = slowsq1(:,:,2:np) - slowsq1(:,:,1:np-1)
      rm = eps*eps*sum(dprod(gradient_m,gradient_m))
      r1 = rd + rm

      slowsq1 = slowsq0 - search * step_size * 1.0
      where(slowsq1 < ssmin) slowsq1 = ssmin
      where(slowsq1 > ssmax) slowsq1 = ssmax
      vel0 = 1./sqrt(slowsq1)
      call model_update(vel0)
      call nlL_oper2(.false., data)
      data = data - data_obs
      rd = sum(dprod(data,data))
      gradient_m(:,:,1) = 0.
      gradient_m(:,:,2:np) = slowsq1(:,:,2:np) - slowsq1(:,:,1:np-1)
      rm = eps*eps*sum(dprod(gradient_m,gradient_m))
      r2 = rd + rm

      slowsq1 = slowsq0 - search * step_size * 1.5
      where(slowsq1 < ssmin) slowsq1 = ssmin
      where(slowsq1 > ssmax) slowsq1 = ssmax
      vel0 = 1./sqrt(slowsq1)
      call model_update(vel0)
      call nlL_oper2(.false., data)
      data = data - data_obs
      rd = sum(dprod(data,data))
      gradient_m(:,:,1) = 0.
      gradient_m(:,:,2:np) = slowsq1(:,:,2:np) - slowsq1(:,:,1:np-1)
      rm = eps*eps*sum(dprod(gradient_m,gradient_m))
      r3 = rd + rm

      write(0,*) "residuals =", real(r0), real(r1), real(r2), real(r3)
      call stepper(r0, r1, r2, r3, step_size, stat)

      if(stat .eq. 0) exit

    end do ! end of trying loop

    if(stat .ne. 0) then
      write(0,*) "trying iterations reached limit without finding a proper step size, will exit the program"
      exit
    endif

    slowsq0 = slowsq0 - step_size * search
    where(slowsq0 < ssmin) slowsq0 = ssmin
    where(slowsq0 > ssmax) slowsq0 = ssmax

    vel0 = 1./sqrt(slowsq0)
    ! Output model
    if(exist_file("mmov")) call sep_write(vel0, "mmov")
    if(exist_file("dmmov")) call sep_write(vel0-vel_init, "dmmov")
    call sep_write(res, "res")
    counter = counter + 1
    if(exist_file("mmov")) call to_history("n4", counter, "mmov")
    if(exist_file("dmmov")) call to_history("n4", counter, "dmmov")
    call to_history("n2", counter, "res")
  end do ! end of iteration loop

  call sep_write(vel0)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
