program FBI

  use sep
  use fbi_plane_mod
  use tomo_mod

  implicit none

  integer,  parameter           :: m = 5, iprint = 0
  double precision, parameter   :: factr  = 1.0d+7, pgtol  = 1.0d-5
  integer                       :: n

  character(len=60)             :: task, csave
  logical                       :: lsave(4)
  integer                       :: isave(44)
  double precision              :: f
  double precision              :: dsave(29)
  integer, allocatable          :: nbd(:), iwa(:)
  double precision, allocatable :: x(:), l(:), u(:), g(:), wa(:)

  integer                             :: i1, i2, i3
  integer                             :: nt, nte, n1, n2, wfld_j, gpu_n, gpu_o
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: bc1, bc2, bc3, bc4, np
  real                                :: dt, d1, o1, d2, o2, op, dp
  real                                :: vmin, vmax, ssmin, ssmax, eps

  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:,:,:), allocatable :: vel_init, vel0, slowsq0, gradient, gradient_m
  real, dimension(:,:,:), allocatable :: data_obs, data

  integer                             :: stat, iter, niter, counter

  double precision                    :: rd, rm

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)
  call from_history("n3", np)
  call from_history("o3", op)
  call from_history("d3", dp)

  ! Read wavelet information
  call from_aux("wavelet", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)

  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  if(gpu_n > np) then
    write(0,*) "reducing the number of GPUs to", np
    gpu_n = np
  endif

  call from_param('niter', niter, 10)
  call from_param('eps', eps)

  call from_param('vmin', vmin)
  call from_param('vmax', vmax)

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, gpu_n, gpu_o, np, op, dp)
  call init4(nte)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)
  call to_history("n3", np)
  call to_history("o3", op)
  call to_history("d3", dp)

  if(exist_file("mmov")) then
    call to_history("n1", n1, "mmov")
    call to_history("o1", o1, "mmov")
    call to_history("d1", d1, "mmov")
    call to_history("n2", n2, "mmov")
    call to_history("o2", o2, "mmov")
    call to_history("d2", d2, "mmov")
    call to_history("n3", np, "mmov")
    call to_history("o3", op, "mmov")
    call to_history("d3", dp, "mmov")
  endif

  if(exist_file("dmmov")) then
    call to_history("n1", n1, "dmmov")
    call to_history("o1", o1, "dmmov")
    call to_history("d1", d1, "dmmov")
    call to_history("n2", n2, "dmmov")
    call to_history("o2", o2, "dmmov")
    call to_history("d2", d2, "dmmov")
    call to_history("n3", np, "dmmov")
    call to_history("o3", op, "dmmov")
    call to_history("d3", dp, "dmmov")
  endif

  call to_history("n1", 3, "res")
  call to_history("o1", 0., "res")
  call to_history("d1", 1., "res")

  n = n1*n2*np

  ! Allocation
  allocate ( nbd(n), x(n), l(n), u(n), g(n) )
  allocate ( iwa(3*n) )
  allocate ( wa(2*m*n + 5*n + 11*m*m + 8*m) )

  allocate( wavelet(nt))
  allocate( res(3))
  allocate( vel0(n1, n2, np))
  allocate( vel_init(n1, n2, np))
  allocate( slowsq0(n1, n2, np))
  allocate( gradient(n1, n2, np))
  allocate( gradient_m(n1, n2, np))
  allocate( data(nte, rec_nx, np))
  allocate( data_obs(nte, rec_nx, np))

  write(0,*) "allocated (3d in inv only) GB", nte/1024.*rec_nx/1024.*np/1024.*4.*2.

  wavelet = 0.
  vel0 = 0.
  vel_init = 0.
  data_obs = 0.
  res = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")
  call sep_read(data_obs, "data")

  where(vel0 < vmin) vel0 = vmin
  where(vel0 > vmax) vel0 = vmax

  ssmax = 1./(vmin*vmin)
  ssmin = 1./(vmax*vmax)

  nbd = 2
  l = ssmin
  u = ssmax

  vel_init = vel0
  slowsq0 = 1./(vel0*vel0)

  do i3=1,np
    do i2=1,n2
      do i1=1,n1
        x(i1+(i2-1)*n1+(i3-1)*n1*n2) = slowsq0(i1,i2,i3)
      enddo
    enddo
  enddo

  call init_alloc(wavelet)

  counter = 0
  iter = 1
  task = 'START'

  do while(task(1:2).eq.'FG'.or.task.eq.'NEW_X'.or.task.eq.'START') 

    call setulb(n,m,x,l,u,nbd,f,g,factr,pgtol,wa,iwa,task,iprint,csave,lsave,isave,dsave)

    if(task(1:2) .eq. 'FG') then

      write(0,*) "starting an FG case"

      do i3=1,np
        do i2=1,n2
          do i1=1,n1
            slowsq0(i1,i2,i3) = x(i1+(i2-1)*n1+(i3-1)*n1*n2)
          enddo
        enddo
      enddo

      vel0 = 1./sqrt(slowsq0)
      call model_update(vel0)

      call nlL_oper2(.false., data)
      data = data - data_obs
      rd = sum(dprod(data,data))

      call L_oper2(.true., .false., gradient, data)

      gradient_m(:,:,1) = 0.
      gradient_m(:,:,2:np) = slowsq0(:,:,2:np) - slowsq0(:,:,1:np-1)
      gradient_m = gradient_m
      rm = eps*eps*sum(dprod(gradient_m,gradient_m))

      gradient_m(:,:,1) = slowsq0(:,:,1) - slowsq0(:,:,2)
      gradient_m(:,:,2:np-1) = 2.*slowsq0(:,:,2:np-1) - slowsq0(:,:,1:np-2) - slowsq0(:,:,3:np)
      gradient_m(:,:,np) = slowsq0(:,:,np) - slowsq0(:,:,np-1)
      gradient_m = eps*eps*gradient_m

      f = rd + rm

      do i3=1,np
        do i2=1,n2
          do i1=1,n1
            g(i1+(i2-1)*n1+(i3-1)*n1*n2) = gradient(i1,i2,i3) + gradient_m(i1,i2,i3)
          enddo
        enddo
      enddo

    else

      if(task(1:5) .eq. 'NEW_X') then   

        write(0,*) "starting a NEW_X case"

        if(isave(34) .ge. niter) task='STOP: TOTAL NO. of f AND g EVALUATIONS EXCEEDS LIMIT'
        if(dsave(13) .le. 1.d-10*(1.0d0 + abs(f))) task='STOP: THE PROJECTED GRADIENT IS SUFFICIENTLY SMALL'

        res(1) = real(rd)
        res(2) = real(rm)
        res(3) = res(1) + res(2)

        write(0,*) "residuals", res

        vel0 = 1./sqrt(slowsq0)
        ! Output model
        if(exist_file("mmov")) call sep_write(vel0, "mmov")
        if(exist_file("dmmov")) call sep_write(vel0-vel_init, "dmmov")
        call sep_write(res, "res")
        counter = counter + 1
        if(exist_file("mmov")) call to_history("n4", counter, "mmov")
        if(exist_file("dmmov")) call to_history("n4", counter, "dmmov")
        call to_history("n2", counter, "res")

        if(task(1:4) .eq. 'STOP') write (0,*) "Stopping: ", task  

      endif
    endif
  enddo

  call sep_write(vel0)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
