program FBI

  use sep
  use fbi_shot_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, gpu_n,  gpu_o
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: bc1, bc2, bc3, bc4
  real                                :: dt, d1, o1, d2, o2

  logical                             :: adj, linear

  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:,:), allocatable :: refl, vel0, data

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("adj", adj)
  if(.not. adj) call from_param("linear", linear)

  if(gpu_n > sou_nx) then
    write(0,*) "reducing the number of GPUs to", sou_nx
    gpu_n = sou_nx
  endif

  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, gpu_n, gpu_o)

  if(adj) then
    call to_history("n1", n1)
    call to_history("o1", o1)
    call to_history("d1", d1)
    call to_history("n2", n2)
    call to_history("o2", o2)
    call to_history("d2", d2)
    call to_history("n3", sou_nx)
    call to_history("o3", sou_ox)
    call to_history("d3", sou_dx)
  else
    call to_history("n1", nt)
    call to_history("o1", 0.)
    call to_history("d1", dt)
    call to_history("n2", rec_nx)
    call to_history("o2", rec_ox)
    call to_history("d2", rec_dx)
    call to_history("n3", sou_nx)
    call to_history("o3", sou_ox)
    call to_history("d3", sou_dx)
  endif

  allocate( wavelet(nt))
  allocate( vel0(n1, n2, sou_nx))
  allocate( refl(n1, n2, sou_nx))
  allocate( data(nt, rec_nx, sou_nx))

  write(0,*) "allocated (3+d in fbi) GB", nt/1024.*rec_nx/1024.*sou_nx/1024.*4. + n1/1024.*n2/1024.*sou_nx/1024.*4.*2.

  wavelet = 0.
  vel0 = 0.
  refl = 0.
  data = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")

  call init_alloc(wavelet)
  call model_update(vel0)

  if(adj) then
    call sep_read(data, "data")
    call L_oper2(adj, .false., refl, data)
    call sep_write(refl)
  else
    if(linear) then
      call sep_read(refl, "refl")
      call L_oper2(adj, .false., refl, data)
      call sep_write(data)
    else
      call nlL_oper2(.false., data)
      call sep_write(data)
    endif
  endif

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
