module fbi_shot_mod

  use sep
  use tomo_mod
  use omp_lib

  implicit none
  integer, private                               :: nt, n1, n2, n3, wfld_j, gpu_n, gpu_o, tid
  integer, private                               :: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x
  integer, private                               :: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x
  integer, private                               :: it
  integer, private                               :: n1e, n2e, bc1, bc2, bc3, bc4, fat
  real, private                                  :: dt, d1, o1, d2, o2, dtw
  real, dimension(:), allocatable, private       :: wavelet, waveletsd
  real, dimension(:,:,:), allocatable, private   :: vel0e, vel0e2, vwt, wt, source, sourcesd, temp_refl, temp_data
  real, dimension(:,:,:,:), allocatable, private :: temp_wfld
  contains

  subroutine init1(n1_in, o1_in, d1_in, n2_in, o2_in, d2_in, bc1_in, bc2_in, bc3_in, bc4_in)
    integer                             :: n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                                :: d1_in, o1_in, d2_in, o2_in
    n1 = n1_in
    n2 = n2_in
    d1 = d1_in
    o1 = o1_in
    d2 = d2_in
    o2 = o2_in
    bc1 = bc1_in
    bc2 = bc2_in
    bc3 = bc3_in
    bc4 = bc4_in
  end subroutine

  subroutine init2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in)
    integer                             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
    integer                             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in
    sou_nx = sou_nx_in
    sou_ox = sou_ox_in
    sou_dx = sou_dx_in
    sou_z = sou_z_in
    rec_nx = rec_nx_in
    rec_ox = rec_ox_in
    rec_dx = rec_dx_in
    rec_z = rec_z_in
  end subroutine

  subroutine init3(nt_in, dt_in, wfld_j_in, gpu_n_in, gpu_o_in)
    integer                             :: nt_in, wfld_j_in, gpu_n_in, gpu_o_in
    real                                :: dt_in
    nt = nt_in
    wfld_j = wfld_j_in
    gpu_n = gpu_n_in
    gpu_o = gpu_o_in
    dt = dt_in
  end subroutine

  subroutine init_alloc(wavelet_in)
    real, dimension(nt)                  :: wavelet_in

    fat = 5
    n1e = n1 + bc1 + bc3 + 2*fat
    n2e = n2 + bc2 + bc4 + 2*fat
    n3 = (nt-1)*wfld_j+1
    dtw = dt / wfld_j

    allocate( wavelet(nt), waveletsd(nt))
    allocate( vel0e(n1e, n2e, sou_nx), vel0e2(n1e, n2e, sou_nx), vwt(n1e, n2e, sou_nx), wt(n1e, n2e, sou_nx))
    allocate( source(nt, sou_nx, gpu_n), sourcesd(nt, sou_nx, gpu_n))
    allocate( temp_refl(n1, n2, gpu_n))
    allocate( temp_data(nt, rec_nx, gpu_n))
    allocate( temp_wfld(n1, n2, nt, gpu_n))

    write(0,*) "allocated (3+d in mod) GB", n1e/1024.*n2e/1024.*sou_nx/1024.*4.*4. + nt/1024.*sou_nx/1024.*gpu_n/1024.*4.*2. + n1/1024.*n2/1024.*gpu_n/1024.*4. + nt/1024.*rec_nx/1024.*gpu_n/1024.*4. + n1/1024.*n2/1024.*nt/1024.*gpu_n*4.

    wavelet = 0.
    waveletsd = 0.

    wavelet = wavelet_in
    call SecDer1d(wavelet, waveletsd)
    waveletsd = - waveletsd / dt / dt

    source = 0.
    sourcesd = 0.

!$omp parallel private(tid) num_threads(gpu_n)
    tid = omp_get_thread_num()+1
    call c_setdevice(tid-1+gpu_o)
    call c_init1(nt, n1, n2, n3, dt, d1, d2)
    call c_init2(n1e, n2e, bc1, bc2, wfld_j)
    call c_init3(sou_nx, sou_ox-1, sou_dx, sou_z-1, rec_nx, rec_ox-1, rec_dx, rec_z-1)
    call c_alloc()
!$omp end parallel

  end subroutine

  subroutine model_update(vel0_in)
    real, dimension(n1, n2, sou_nx)              :: vel0_in
    vel0e = 0.
    vwt = 0.
    wt = 0.
    vel0e(1+fat+bc1:n1+fat+bc1,1+fat+bc2:n2+fat+bc2,:) = vel0_in
    do isou=1,sou_nx
      call extend(vel0e(:,:,isou), bc1+fat, bc2+fat, bc3+fat, bc4+fat)
      call AbcWt(vel0e(:,:,isou), vwt(:,:,isou), wt(:,:,isou), dtw, d1, d2, bc1, bc2, bc3, bc4, fat)
    enddo
    vel0e2 = vel0e*vel0e*dtw*dtw
  end subroutine

  function nlL_oper(add, d) result(stat)
    integer            :: stat
    logical,intent(in) :: add 
    real,dimension(:)  :: d
    call nlL_oper2(add,d)
    stat=0
  end function

  function L_oper(adj, add, m, d) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj, add 
    real,dimension(:)  :: m,d
    call L_oper2(adj,add,m,d)
    stat=0
  end function

  subroutine nlL_oper2(add, data)
    logical, intent(in)                  :: add
    real, dimension(nt, rec_nx, sou_nx)  :: data
    if(.not. add) data = 0.
!$omp parallel private(tid, isou) num_threads(gpu_n)
    tid = omp_get_thread_num()+1
    call c_setdevice(tid-1+gpu_o)
!$omp do schedule(dynamic, 1)
    do isou=1,sou_nx
      call c_model_update(vel0e2(:,:,isou), vwt(:,:,isou), wt(:,:,isou))
      source(:,:,tid) = 0.
      source(:,isou,tid) = wavelet
      temp_data(:,:,tid) = 0.
      call c_nlmodel(source(:,:,tid), temp_data(:,:,tid))
      data(:,:,isou) = data(:,:,isou) + temp_data(:,:,tid)
    enddo
!$omp end do
!$omp end parallel

  end subroutine

  subroutine L_oper2(adj, add, refl, data)
    logical, intent(in)                  :: adj, add
    real, dimension(n1, n2, sou_nx)          :: refl
    real, dimension(nt, rec_nx, sou_nx)     :: data
    if(adj) then
      if(.not. add) refl = 0.
!$omp parallel private(tid, isou) num_threads(gpu_n)
    tid = omp_get_thread_num()+1
    call c_setdevice(tid-1+gpu_o)
!$omp do schedule(dynamic, 1)
      do isou=1,sou_nx
        call c_model_update(vel0e2(:,:,isou), vwt(:,:,isou), wt(:,:,isou))
        sourcesd(:,:,tid) = 0.
        sourcesd(:,isou,tid) = waveletsd
        temp_wfld(:,:,:,tid) = 0.
        temp_refl(:,:,tid) = 0.
        call c_migrate(sourcesd(:,:,tid), temp_wfld(:,:,:,tid), temp_refl(:,:,tid), data(:,:,isou))
        refl(:,:,isou) = refl(:,:,isou) + temp_refl(:,:,tid)
      enddo
!$omp end do
!$omp end parallel
    else
      if(.not. add) data = 0.
!$omp parallel private(tid, isou) num_threads(gpu_n)
    tid = omp_get_thread_num()+1
    call c_setdevice(tid-1+gpu_o)
!$omp do schedule(dynamic, 1)
      do isou=1,sou_nx
        call c_model_update(vel0e2(:,:,isou), vwt(:,:,isou), wt(:,:,isou))
        sourcesd(:,:,tid) = 0.
        sourcesd(:,isou,tid) = waveletsd
        temp_data(:,:,tid) = 0.
        call c_model(sourcesd(:,:,tid), refl(:,:,isou), temp_data(:,:,tid))
        data(:,:,isou) = data(:,:,isou) + temp_data(:,:,tid)
      enddo
!$omp end do
!$omp end parallel
    endif
  end subroutine

end module 
